#############################################################################
##
#W  capable.gi            share package 'enumpgrp'              Bettina Eick
##

#############################################################################
##
#F PCoveringGroup( G )
##
PCoveringGroup := function( G )
    local Gcgs, n, r, p, c, Ggens, F, Fgens, hom, def, i, Frels, w, v, j, 
          Prels, f, H;

    # set up
    Gcgs  := SpecialPcgs( G );
    n     := Length( Gcgs );
    r     := RankPGroup( G );
    p     := PrimePGroup( G );
    c     := PClassPGroup( G );
    Ggens := Gcgs{[1..r]};
  
    # compute presenation on minimal generating set
    F     := FreeGroup( r );
    Fgens := GeneratorsOfGroup( F );
    hom   := GroupHomomorphismByImages( F, G, Fgens, Ggens );
    SetIsSurjective( hom, true );

    # get definitions of cgs
    def := [];
    for i in [1..r] do
        def[i] := Fgens[i];
    od;
    for i in [r+1..n] do
        def[i] := PreImagesRepresentative( hom, Gcgs[i] );
    od;

    # compute relators for presentation
    Frels := [];
    for i in [1..n] do
        w := def[i]^p;
        v := PreImagesRepresentative( hom, Gcgs[i]^p );
        Add( Frels, w / v );
    od;
    for i in [1..n] do
        for j in [i+1..n] do
            w := Comm( def[j], def[i] );        
            v := Comm( Gcgs[j], Gcgs[i] );
            v := PreImagesRepresentative( hom, v );
            Add( Frels, w / v );
        od;
    od;

    # now compute relators for P
    Prels := [];
    for f in Fgens do
        Append( Prels, List( Frels, x -> Comm(x, f) ) );
    od;
    Append( Prels, List( Frels, x -> x^p ) );

    H := F/Prels;
    H := PQuotient( H, p, c+1);
    return Image( EpimorphismQuotientSystem( H ) );
end;
 
#############################################################################
##
#F Capability( G )
##
Capability := function( G )
    local P, c, ser;
    P := PCoveringGroup(G);
    c := PClassPGroup(G);
    ser := PCentralSeries( P, PrimePGroup( G ) );
    return Length( Pcgs( ser[c+1] ) );
end;
