#############################################################################
##
#W  enumall.gi             share package 'enumpgrp'              Bettina Eick
##

#############################################################################
##
#F AddConjugacyClassesModule( M )
##
AddConjugacyClassesModule := function( M )
    local p, d;
    if not IsBound( M.classes ) then
        d := M.dimension;
        p := Characteristic( M.field );
        M.classes := ConjugacyClasses( GL( d, p ) );
    fi;
    return M.classes;
end;

#############################################################################
##
#F EnumerateDescendantsWithClasses2( G, M, clA, clB ) 
##
EnumerateDescendantsWithClasses2 := function( G, M, clA, clB )
    local cc, p, gens, mats, id, sub, s, t, i, j, l, mat, null, r, u, ints,
          n, full, d, gensA, gensB, f, perm, A, B, tabi, tabs, pos, h;

    # set up
    A := ActingDomain(clA[1]);
    B := ActingDomain(clB[1]);

    # compute cohomology
    Info(InfoEnum, 2, "  compute cohomology ");
    cc := TwoCohomology( G, M );
    p  := Characteristic( M.field );
    d  := M.dimension;
    h  := Dimension( Source( cc.cohom ) ) / d;

    # compute matrix action on cc
    Info(InfoEnum, 2, "  compute matrix action ");
    gensA := List( clA, x -> Tuple( [Representative(x), Identity( B ) ] ) );
    gensB := List( clB, x -> Tuple( [Identity(A), Representative(x)] ) );
    mats := MatrixOperationOfCPGroupWT( cc, Concatenation( gensA, gensB ) );
    mats := List( mats, TransposedMat );
    id := mats[1]^0;
    n := Length( clA );

    # the non-descendant subspaces
    Info(InfoEnum, 2, "  compute non-descendant subspaces ");
    sub := NonDescendantSubspaces( G, M, cc );
    Sort( sub );

    # set up look up table
    tabi := [];
    tabs := [];

    # loop over conjugacy classes
    Info(InfoEnum, 2, "  start counting fixed points ");
    s := 0;
    for i in [1..Length(clA)] do
        for j in [1..Length(clB)] do
            l := Size(clA[i]) * Size( clB[j] );
            mat := mats[i] * mats[n+j];

            null := NullspaceMat( mat - id );
            null := VectorSpace( M.field, null, Zero(Image(cc.cohom)));
            full := IsFull( p, d, h, null );

            if IsFlat( p, d, h, null ) then
                r := 0;
                Info( InfoEnum, 3, "  flat sp ... 0 fixed points");
            elif not IsBool( full ) then
                r := SizeOfFullUnion( p, d, full);
                Info( InfoEnum, 3, "  full sp ... ",r," fixed points");
            else
                r := Size( null );
 
                # compute intersections
                Info(InfoEnum, 4, "  compute intersections ");
                ints := List(sub, x -> IntersectionVS(x, null));

                Info(InfoEnum, 4, "  find maximals ");
                ints := FindMaximals( ints );
                Sort( ints );

                # first try to look up
                pos := Position( tabi, ints );
                if not IsBool( pos ) then
                    r := r - tabs[pos];
                    Info( InfoEnum, 3, "  look up ... ",r," fixed points");
                else
                    Info(InfoEnum, 4, "  compute unions with ",
                                  List(ints, Dimension) );
                    u := SizeOfUnion( ints );
                    r := r - u;
                    Info( InfoEnum, 3, "  unions  ... ",r," fixed points");
                    Add( tabi, ints );
                    Add( tabs, u );
                fi;
            fi;
            s := s + r*l;
        od;
    od; 

    # return
    return s/(Size(A)*Size(B));
end;

#############################################################################
##
#F EnumerateDescendantsWithClasses( G, M, clA, clB ) 
##
EnumerateDescendantsWithClasses := function( G, M, clA, clB )
    local cc, p, gens, mats, id, sub, s, t, i, j, l, mat, null, r, u, ints,
          n, full, d, gensA, gensB, f, perm, A, B, tabi, tabs, pos, h, V;

    # set up
    A := ActingDomain(clA[1]);
    B := ActingDomain(clB[1]);

    # compute cohomology
    Info(InfoEnum, 2, "  compute cohomology ");
    cc := TwoCohomology( G, M );
    V := Image( cc.cohom );

    # compute matrix action on cc
    Info(InfoEnum, 2, "  compute matrix action ");
    gensA := List( clA, x -> Tuple( [Representative(x), Identity( B ) ] ) );
    gensB := List( clB, x -> Tuple( [Identity(A), Representative(x)] ) );
    mats := MatrixOperationOfCPGroupWT( cc, Concatenation( gensA, gensB ) );
    mats := List( mats, TransposedMat );
    id := mats[1]^0;
    n := Length( clA );

    # the non-descendant subspaces
    Info(InfoEnum, 2, "  compute non-descendant subspaces ");
    sub := NonDescendantSubspaces( G, M, cc );
    Sort( sub );

    # initialize look up table
    tabs := [];
    tabi := [];

    # loop over conjugacy classes
    Info(InfoEnum, 2, "  start counting fixed points ");
    s := 0;
    for i in [1..Length(clA)] do
        for j in [1..Length(clB)] do
            l := Size(clA[i]) * Size( clB[j] );
            mat := TransposedMat(mats[i] * mats[n+j]);  # -- wieso Trsps?

            null := NullspaceMat( mat - id );
            null := VectorSpace(M.field, null, Zero(V));
            null := IntersectionVS( V, null );

            if Dimension( null ) = 0 then
                r := 0;
            else
                r := Size( null );
 
                # compute intersections
                Info(InfoEnum, 4, "  compute intersections ");
                ints := List(sub, x->IntersectionVS(null, x));
                Info(InfoEnum, 4, "  find maximals ");
                ints := FindMaximals( ints );
                
                # try to look up
                pos := Position( tabi, ints );
                if not IsBool( pos ) then
                    r := r - tabs[pos];
                    Info( InfoEnum, 3, "  look up ... ",r," fixed points");
                else
                    Info(InfoEnum, 4, "  compute unions with ", 
                              List( ints, Dimension ));
                    u := SizeOfUnion( ints );
                    r := r - u;
                    Info( InfoEnum, 3, "  unions  ... ",r," fixed points");
                    Add( tabi, ints );
                    Add( tabs, u );
                fi;
            fi;
            if r < 0 then Error("found strange size "); fi;
            s := s + r*l;
        od;
    od; 

    # return
    return s/(Size(A)*Size(B));
end;

#############################################################################
##
#F EnumerateDescendants( G, M ) 
##
EnumerateDescendants := function( G, M )
    local A, clA, clB; 

    # the classes of Aut(M)
    Info(InfoEnum, 2, "  add classes of Aut(M) ");
    AddConjugacyClassesModule( M );
    clB := M.classes;

    # the classes of Aut(G)
    Info(InfoEnum, 2, "  compute classes of Aut(G) ");
    A := AutomorphismGroup( G );
    clA := ConjugacyClassesAutGrp( A );

    # enumerate
    if IsElementaryAbelian(G) then
        return  EnumerateDescendantsWithClasses2( G, M, clA, clB );
    elif Capability( G ) >= M.dimension then
        return  EnumerateDescendantsWithClasses( G, M, clA, clB );
    else
        return 0;
    fi;
end;

#############################################################################
##
#F EnumerateGroupsByStepsize( p, l, d )
##
EnumerateGroupsByStepsize := function( p, l, d )
    local c,  M, n, i, G, A, clA, e;

    # get the module
    M := TrivialGModuleOfGroup( SmallGroup( p^l, 1 ), d );
    AddConjugacyClassesModule( M );

    # descendants of the cyclic group
    if d = 1 then 
        c := 1;
    else
        c := 0;
    fi;
    if l = 1 then return c; fi;

    # descendants of non-trivial groups
    n := NumberSmallGroups( p^l );
    for i in [2..n-1] do
        G := SmallGroup( p^l, i );
        Info( InfoEnum, 1, "start group ",p^l," - ",i);
        if Capability( G ) >= d then
            IsPGroup(G);
            A := AutomorphismGroup(G);
            clA := ConjugacyClassesAutGrp( A );
            e := EnumerateDescendantsWithClasses( G, M, clA, M.classes );
            c := c + e;
        fi;
    od;

    # descendants of the elementary abelian group
    Info( InfoEnum, 1, "start group ",p^l," - final");
    if d = l*(l+1)/2 then return c+1; fi;
    if d <= l*(l+1)/2 then 
        G := ElementaryAbelianGroup( p^l );
        A := AutomorphismGroup( G );
        clA := ConjugacyClasses( GL( l, p ) );
        clA := TranslateMatsToAutomorphisms( G, A, clA );
        e := EnumerateDescendantsWithClasses2( G, M, clA, M.classes );
        c := c + e;
    fi;

    return c;
end;

#############################################################################
##
#F EnumerateGroupsByOrder( p, n )
##
EnumerateGroupsByOrder := function( p, n )
    local c, e, d;

    # check args
    if not IsPrime( p ) then return fail; fi;
    if not IsInt( n ) and n > 0 then return fail; fi;

    # the trivial case
    if n = 1 then return 1; fi;

    # stepsize 0 yields the elementary abelian group
    c := 1;

    # the other stepsizes
    for d in [1..n-1] do
        e := EnumerateGroupsByStepsize( p, n-d, d );
        c := c + e;
    od;
    return c;
end;

